"""A getter for HGNC."""

import logging
from typing import ClassVar

import requests

from bioversions.utils import Getter, VersionType

__all__ = [
    "HGNCGetter",
]

logger = logging.getLogger(__name__)

PATH = "https://storage.googleapis.com/public-download-files/hgnc/archive/archive/monthly/json/"
URL = "https://storage.googleapis.com/storage/v1/b/public-download-files/o?prefix=hgnc/archive/archive/monthly"
PREFIX = "hgnc/archive/archive/monthly/json/hgnc_complete_set_"
SUFFIX = ".json"


class HGNCGetter(Getter):
    """A getter for HGNC."""

    collection: ClassVar[list[str]] = ["hgnc", "hgnc.symbol", "hgnc.genegroup", "hgnc.genesymbol"]
    name = "HGNC"
    version_type = VersionType.date

    def get(self) -> str:
        """Get the latest monthly HGNC version number."""
        res = requests.get(URL, timeout=5)
        items = res.json()["items"]
        return max(
            item["name"].removeprefix(PREFIX).removesuffix(SUFFIX)
            for item in items
            if (name := item["name"]).startswith(PREFIX) and name.endswith(SUFFIX)
        )


if __name__ == "__main__":
    HGNCGetter.print()



"""Command line interface for :mod:`bioversions`."""

import click
from click_default_group import DefaultGroup
from more_click import make_web_command, verbose_option
from tabulate import tabulate

from bioversions.resources.update import update


@click.group(cls=DefaultGroup, default="web", default_if_no_args=True)
@click.version_option()
def main() -> None:
    """The bioversions CLI."""  # noqa:D401


main.add_command(update)

web = make_web_command(
    app="bioversions.wsgi:app",
    group=main,
    command_kwargs={
        "help": "Run the bioversions web application.",
    },
)


@main.command()  # type:ignore
@click.argument("key")
@verbose_option  # type:ignore
def get(key: str) -> None:
    """Print the version."""
    from . import get_version

    click.echo(get_version(key))


@main.command()  # type:ignore
@click.option("--terse", "-t", is_flag=True)
def ls(terse: bool) -> None:
    """List versions."""
    from . import get_rows

    if terse:
        click.echo(
            tabulate(
                sorted(
                    (bv.bioregistry_id or "", bv.classname, bv.version)
                    for bv in get_rows(use_tqdm=True)
                ),
                headers=["Prefix", "Class", "Version"],
            )
        )
    else:
        click.echo(
            tabulate(
                (
                    (bv.bioregistry_id, bv.name, bv.version, bv.date, bv.homepage)
                    for bv in get_rows(use_tqdm=True)
                ),
                headers=["Prefix", "Name", "Version", "Release Date", "Homepage"],
            )
        )


if __name__ == "__main__":
    main()



"""A getter for UMLS."""

from typing import ClassVar

from bioversions.utils import Getter, VersionType, find, get_soup

__all__ = [
    "UMLSGetter",
]

URL = "https://www.nlm.nih.gov/research/umls/licensedcontent/umlsknowledgesources.html"


class UMLSGetter(Getter):
    """A getter for UMLS."""

    collection: ClassVar[list[str]] = ["umls", "sty", "umls.aui"]
    name = "UMLS"
    version_type = VersionType.other

    def get(self) -> str:
        """Get the latest UMLS version number."""
        soup = get_soup(URL)
        main_tag = find(soup, "main")
        t1 = find(main_tag, "div", {"class": "grid-row grid-gap-1"})  # type:ignore
        t2 = find(t1, "div", {"class": "tablet:grid-col-12"})
        version_tag = find(t2, "h2")
        version = version_tag.text.split()[0]
        return version


if __name__ == "__main__":
    UMLSGetter.print()



"""A getter for LOINC."""

from bioversions.utils import Getter, VersionType, get_soup

__all__ = ["LOINCGetter"]


class LOINCGetter(Getter):
    """A getter for LOINC."""

    bioregistry_id = "loinc"
    name = "Logical Observation Identifiers Names and Codes"
    version_type = VersionType.semver_minor
    date_fmt = "%Y-%m-%d"

    def get(self) -> dict[str, str]:
        """Get the latest LOINC version number."""
        soup = get_soup("https://loinc.org")
        current_version_div = soup.find(id="current-version")
        if current_version_div is None:
            raise ValueError

        h3 = current_version_div.find("h3")
        if h3 is None or not h3.text:
            raise ValueError

        version = h3.text.removeprefix("LOINC Version ")

        paragraph = current_version_div.find("p")
        if paragraph is None or not paragraph.text:
            raise ValueError

        date = paragraph.text.split("Released ")[1].strip()
        return {"version": version, "date": date}


if __name__ == "__main__":
    LOINCGetter.print()



"""Sources for Bioversions."""

from __future__ import annotations

import ftplib
import traceback
import warnings
from collections.abc import Iterable
from typing import Literal, NamedTuple, overload

from class_resolver import ClassResolver
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from .antibodyregistry import AntibodyRegistryGetter
from .bigg import BiGGGetter
from .biogrid import BioGRIDGetter
from .ccle import CCLEGetter
from .cellosaurus import CellosaurusGetter
from .chebi import ChEBIGetter
from .chembl import ChEMBLGetter
from .chemidplus import ChemIDplusGetter
from .civic import CiVICGetter
from .complexportal import ComplexPortalGetter
from .daily import NCBIGeneGetter
from .depmap import DepMapGetter
from .dgi import DGIGetter
from .disgenet import DisGeNetGetter
from .drugbank import DrugBankGetter
from .drugcentral import DrugCentralGetter
from .ensembl import EnsemblGetter
from .expasy import ExPASyGetter
from .flybase import FlybaseGetter
from .gtdb import GTDBGetter
from .guidetopharmacology import GuideToPharmacologyGetter
from .hgnc import HGNCGetter
from .homologene import HomoloGeneGetter
from .icd10 import ICD10Getter
from .icd11 import ICD11Getter
from .icf import ICFGetter
from .intact import IntActGetter
from .interpro import InterProGetter
from .itis import ITISGetter
from .kegg import KEGGGetter
from .loinc import LOINCGetter
from .mesh import MeshGetter
from .mgi import MGIGetter
from .mirbase import MirbaseGetter
from .moalmanac import MOAlmanacGetter
from .msigdb import MSigDBGetter
from .ncit import NCItGetter
from .npass import NPASSGetter
from .obo import DoidGetter, GoGetter
from .ols import extend_ols
from .omim import OMIMGetter
from .oncotree import OncoTreeGetter
from .pathbank import PathBankGetter
from .pathwaycommons import PathwayCommonsGetter
from .pfam import PfamGetter
from .pombase import PombaseGetter
from .pr import PRGetter
from .pubchem import PubChemGetter
from .reactome import ReactomeGetter
from .rfam import RfamGetter
from .rgd import RGDGetter
from .rhea import RheaGetter
from .ror import RORGetter
from .rxnorm import RxNormGetter
from .sgd import SgdGetter
from .signor import SignorGetter
from .silva import SILVAGetter
from .slm import SwissLipidGetter
from .spdx import SPDXGetter
from .stringdb import StringDBGetter
from .umls import UMLSGetter
from .uniprot import UniProtGetter
from .wikipathways import WikiPathwaysGetter
from .zfin import ZfinGetter
from ..utils import (
    DailyGetter,
    Getter,
    OBOFoundryGetter,
    UnversionedGetter,
    VersionResult,
    refresh_daily,
)

__all__ = [
    "AntibodyRegistryGetter",
    "BiGGGetter",
    "BioGRIDGetter",
    "CCLEGetter",
    "CellosaurusGetter",
    "ChEBIGetter",
    "ChEMBLGetter",
    "ChemIDplusGetter",
    "CiVICGetter",
    "ComplexPortalGetter",
    "DGIGetter",
    "DepMapGetter",
    "DisGeNetGetter",
    "DoidGetter",
    "DrugBankGetter",
    "DrugCentralGetter",
    "EnsemblGetter",
    "ExPASyGetter",
    "FlybaseGetter",
    "GTDBGetter",
    "GoGetter",
    "GuideToPharmacologyGetter",
    "HGNCGetter",
    "HomoloGeneGetter",
    "ICD10Getter",
    "ICD11Getter",
    "ICFGetter",
    "ITISGetter",
    "IntActGetter",
    "InterProGetter",
    "KEGGGetter",
    "LOINCGetter",
    "MGIGetter",
    "MOAlmanacGetter",
    "MSigDBGetter",
    "MeshGetter",
    "MirbaseGetter",
    "NCBIGeneGetter",
    "NCItGetter",
    "NPASSGetter",
    "OMIMGetter",
    "OncoTreeGetter",
    "PRGetter",
    "PathBankGetter",
    "PathwayCommonsGetter",
    "PfamGetter",
    "PombaseGetter",
    "PubChemGetter",
    "RGDGetter",
    "RORGetter",
    "ReactomeGetter",
    "RfamGetter",
    "RheaGetter",
    "RxNormGetter",
    "SILVAGetter",
    "SPDXGetter",
    "SgdGetter",
    "SignorGetter",
    "StringDBGetter",
    "SwissLipidGetter",
    "UMLSGetter",
    "UniProtGetter",
    "VersionFailure",
    "WikiPathwaysGetter",
    "ZfinGetter",
    "clear_cache",
    "get_rows",
    "get_version",
    "getter_resolver",
    "iter_versions",
    "resolve",
]

#: These are broken beyond fixing at the moment
SKIPPED = {
    DrugBankGetter,
    PathwayCommonsGetter,
    DisGeNetGetter,
    # Upper-level classes
    OBOFoundryGetter,
    UnversionedGetter,
    DailyGetter,
}

getter_resolver: ClassResolver[Getter] = ClassResolver.from_subclasses(
    base=Getter,
    suffix="Getter",
    skip=SKIPPED,
    synonym_attribute=["collection"],
)
extend_ols(getter_resolver)


def get_getters() -> list[type[Getter]]:
    """Get a list of getters."""
    warnings.warn("iterate over getter_resolver directly", DeprecationWarning, stacklevel=2)
    return list(getter_resolver)


def resolve(name: str | type[Getter], *, use_cache: bool = True) -> VersionResult:
    """Resolve the database name to a :class:`Bioversion` instance."""
    if use_cache:
        if isinstance(name, type):
            name = name.__name__
        return _resolve_helper_cached(name)
    else:
        if isinstance(name, str):
            name = getter_resolver.lookup(name)
        return name.resolve()


@refresh_daily
def _resolve_helper_cached(name: str) -> VersionResult:
    getter = getter_resolver.lookup(name)
    return getter.resolve()


def clear_cache() -> None:
    """Clear the cache."""
    _resolve_helper_cached.clear_cache()


# docstr-coverage:excused `overload`
@overload
def get_version(name: str, *, strict: Literal[True] = True) -> str: ...


# docstr-coverage:excused `overload`
@overload
def get_version(name: str, *, strict: Literal[False] = False) -> str | None: ...


def get_version(name: str, *, strict: bool = True) -> str | None:
    """Resolve a database name to its version string.

    :param name:
        The name of the resource to get the version from. Often, this is a Bioregistry
        prefix, but sometimes can be an ad-hoc key for a database.
    :param strict:
        Re-raises errors in version resolution by default. Set explicitly to
        ``false`` to return None on errors.
    :return: The version of the resource as a string
    """
    try:
        rv = resolve(name).version
    except Exception:
        if strict:
            raise
        return None
    else:
        return rv


def get_rows(use_tqdm: bool | None = False) -> list[VersionResult]:
    """Get the rows, refreshing once per day."""
    return [
        bioversion
        for bioversion in iter_versions(use_tqdm=use_tqdm)
        if isinstance(bioversion, VersionResult)
    ]


class VersionFailure(NamedTuple):
    """Holds information about failures."""

    name: str
    clstype: str
    message: str
    trace: str


def iter_versions(
    use_tqdm: bool | None = False,
) -> Iterable[VersionResult | VersionFailure]:
    """Iterate over versions, without caching."""
    with logging_redirect_tqdm():
        getters = tqdm(
            list(getter_resolver), disable=not use_tqdm, desc="Getting versions", unit="resource"
        )
        for cls in getters:
            getters.set_postfix(name=cls.name)
            try:
                yv = resolve(cls)
            except (OSError, AttributeError, ftplib.error_perm) as e:
                msg = f"[{cls.bioregistry_id or cls.name}] failed to resolve: {e}"
                tqdm.write(msg)
                yield VersionFailure(cls.name, cls.__name__, msg, traceback.format_exc())
            except (ValueError, KeyError) as e:
                msg = f"[{cls.bioregistry_id or cls.name}] issue parsing: {e}"
                tqdm.write(msg)
                yield VersionFailure(cls.name, cls.__name__, msg, traceback.format_exc())
            else:
                yield yv



"""A getter for OncoTree."""

import requests

from ..utils import Getter, VersionType

__all__ = [
    "OncoTreeGetter",
]


class OncoTreeGetter(Getter):
    """A getter for OncoTree."""

    bioregistry_id = "oncotree"
    name = "OncoTree"
    homepage_fmt = "http://oncotree.mskcc.org/api/tumorTypes?version=oncotree_{version}"
    version_type = VersionType.date
    date_version_fmt = "%Y-%m-%d"

    def get(self) -> str:
        """Get the latest OncoTree version number."""
        res = requests.get(
            "http://oncotree.mskcc.org/api/versions", params={"format": "json"}, timeout=5
        )
        res_json = res.json()
        version = next(
            r["release_date"] for r in res_json if r["api_identifier"] == "oncotree_latest_stable"
        )
        return version

    @staticmethod
    def homepage_version_transform(version: str) -> str:
        """Reformat version value for OncoTree data URL."""
        return version.replace("-", "_")


if __name__ == "__main__":
    OncoTreeGetter.print()



"""A getter for miRBase."""

from typing import ClassVar

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "MirbaseGetter",
]

PREFIX = "0_THIS_IS_RELEASE_"


class MirbaseGetter(Getter):
    """A getter for miRBase."""

    bioregistry_id = "mirbase"
    name = "miRBase"
    homepage_fmt = "https://www.mirbase.org/download/PREVIOUS_RELEASES/{version}"
    version_type = VersionType.semver_minor
    collection: ClassVar[list[str]] = ["mirbase", "mirbase.family", "mirbase.mature"]

    def get(self):
        """Get the latest miRBase version number."""
        return "22.1"


def _old():
    """Get miRBase version number from the old server."""
    # Old code, doesn't work anymore because miRBase site deleted
    # most of their FTP server and downloads
    url = "http://www.mirbase.org/download_readme/"
    soup = get_soup(url, verify=False)
    return soup.find("p").text.splitlines()[0].split()[-1]


if __name__ == "__main__":
    MirbaseGetter.print()



"""A getter for Rfam."""

import requests

from bioversions.utils import Getter, VersionType

__all__ = [
    "RfamGetter",
]


class RfamGetter(Getter):
    """A getter for Rfam."""

    bioregistry_id = "rfam"
    name = "Rfam"
    homepage_fmt = "https://ftp.ebi.ac.uk/pub/databases/Rfam/{version}/"
    version_type = VersionType.semver_minor

    def get(self) -> str:
        """Get the latest Rfam version number."""
        res = requests.get("https://ftp.ebi.ac.uk/pub/databases/Rfam/CURRENT/README", timeout=15)
        for line in res.iter_lines():
            line = line.decode("utf8").strip()
            if line.startswith("Release "):
                return line[len("Release ") :]
        raise ValueError


if __name__ == "__main__":
    RfamGetter.print()



"""A getter for FlyBase."""

import re

from bioversions.utils import Getter, VersionType, get_soup

__all__ = [
    "FlybaseGetter",
]

URL = "https://s3ftp.flybase.org/releases/"
PATTERN = re.compile(r"FB\d{4}_\d{2}")
AGENT = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
    "(KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36"
)


class FlybaseGetter(Getter):
    """A getter for FlyBase."""

    bioregistry_id = "flybase"
    name = "FlyBase"
    homepage_fmt = "https://s3ftp.flybase.org/releases/FB{version}/"
    version_type = VersionType.month

    def get(self) -> str:
        """Get the latest FlyBase version."""
        soup = get_soup(URL, user_agent=AGENT)
        releases = [
            match.group().removeprefix("FB")
            for anchor_tag in soup.find_all("a", href=True)
            if (match := PATTERN.search(anchor_tag.text))
        ]
        if not releases:
            raise ValueError("flybase hit anti-scraping measurements")
        latest_version = max(releases)
        return latest_version


if __name__ == "__main__":
    FlybaseGetter.print()



"""A getter for ChEMBL."""

import ftplib
import io
from typing import ClassVar

from bioversions.utils import Getter, VersionType

__all__ = [
    "ChEMBLGetter",
]

RELEASE_PREFIX = "* Release:"
DATE_PREFIX = "* Date:"


class ChEMBLGetter(Getter):
    """A getter for ChEMBL."""

    bioregistry_id = "chembl"
    name = "ChEMBL"
    homepage_fmt = "ftp://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_{version}"
    date_fmt = "%d/%m/%Y"
    version_type = VersionType.sequential
    collection: ClassVar[list[str]] = ["chembl", "chembl.target", "chembl.compound", "chembl.cell"]

    def get(self):
        """Get the latest ChEMBL version number."""
        bio = io.BytesIO()
        with ftplib.FTP("ftp.ebi.ac.uk") as ftp:
            ftp.login()
            ftp.retrbinary("RETR pub/databases/chembl/ChEMBLdb/latest/README", bio.write)
        bio.seek(0)
        version, date = None, None
        for line in bio.read().decode("utf-8").split("\n"):
            if line.startswith(RELEASE_PREFIX):
                version = _removeprefix(_removeprefix(line, RELEASE_PREFIX).strip(), "chembl_")
            elif line.startswith(DATE_PREFIX):
                date = _removeprefix(line, DATE_PREFIX).strip()
        if version is None or date is None:
            raise ValueError
        return {"date": date, "version": version}


def _removeprefix(s, prefix):
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


if __name__ == "__main__":
    ChEMBLGetter.print()



"""Getters for OBO ontologies."""

from bioversions.utils import OBOFoundryGetter, VersionType

__all__ = [
    "DoidGetter",
    "GoGetter",
]


class GoGetter(OBOFoundryGetter):
    """A getter for the Gene Ontology (GO)."""

    name = "Gene Ontology"
    bioregistry_id = "go"
    version_type = VersionType.date
    strip_version_prefix = True
    date_version_fmt = "%Y-%m-%d"
    homepage_fmt = "http://archive.geneontology.org/full/{version}/"


class DoidGetter(OBOFoundryGetter):
    """A getter for the Disease Ontology (DO)."""

    name = "Disease Ontology"
    homepage_fmt = "https://github.com/DiseaseOntology/HumanDiseaseOntology/tree/main/src/ontology/releases/{version}"
    bioregistry_id = "doid"
    version_type = VersionType.date
    strip_version_prefix = True
    strip_file_suffix = True
    date_version_fmt = "%Y-%m-%d"



"""A getter for ICD11."""

import warnings

import requests
from urllib3.exceptions import InsecureRequestWarning

from bioversions.utils import Getter, VersionType

__all__ = [
    "ICD11Getter",
]

URL = "https://icd.who.int/browse/latest-release/mms/en"


class ICD11Getter(Getter):
    """A getter for ICD11."""

    # TODO combine with ICF using `collection`
    bioregistry_id = "icd11"
    name = "International Classification of Diseases, 11th Revision"
    version_type = VersionType.date
    date_version_fmt = "%Y-%m"

    def get(self) -> str:
        """Get the latest ICD11 version number."""
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=InsecureRequestWarning)
            response = requests.get(URL, allow_redirects=True, timeout=15, verify=False)  # noqa:S501
        final_url = response.url
        return final_url[len("https://icd.who.int/browse/") :].split("/")[0]


if __name__ == "__main__":
    ICD11Getter.print()



"""A getter for Pathway Commons."""

from bioversions.utils import Getter, VersionType, find, get_soup

__all__ = [
    "PathwayCommonsGetter",
]

URL = "https://www.pathwaycommons.org/"


class PathwayCommonsGetter(Getter):
    """A getter for Pathway Commons."""

    name = "Pathway Commons"
    version_type = VersionType.sequential

    def get(self) -> str:
        """Get the latest Pathway Commons version number."""
        soup = get_soup(URL)
        boost = find(soup, {"class": "boost"})
        boost_text = boost.text[len("Version ") :]
        boost_text = boost_text.split(":")[0]
        return boost_text


if __name__ == "__main__":
    PathwayCommonsGetter.print()
